/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Analyzer.scala 2549 2006-01-11 18:05:40 +0100 (Wed, 11 Jan 2006) mihaylov $
 */

package zweic;

/**
 * This class implements the semantic analyzer of zweic.
 */
class Analyzer {

  import Operators._;
  import scala.collection.immutable.{Map, ListMap};

  // Global scope of classes
  type ClassScope = Map[String, ClassSymbol];
  var classScope: ClassScope = ListMap.Empty;

  // For local scopes of variables
  type VarScope = Map[String, VarSymbol];
  val emptyVarScope: VarScope = ListMap.Empty;

  // Analyze a program
  def analyzeProgram(tree: Program): Unit =
    tree match {

      case Program(classes, main) =>
        classes.foreach(analyzeClass);
        analyzeExpr(emptyVarScope, main);
        ()
    }

  // Analyze a class
  private def analyzeClass(tree: ClassDef): Unit =
    // ...  complter ...

  // Analyze a member
  private def analyzeMember(ownerClass: ClassSymbol, tree: Member): Unit =
    // ...  complter ...

  // Analyze a statement
  private def analyzeStat(varScope: VarScope, tree: Stat): VarScope =
    tree match {

      case While(cond, stats) =>
        checkIntType(cond.pos, analyzeExpr(varScope, cond));
        stats.foldLeft(varScope)(analyzeStat);
        varScope
      // ...  complter ...
    }

  // Analyze a type
  private def analyzeType(tree: TypeTree): Type =
    tree match {
      case IntType() => IIntType
      case NullType() => INullType
      case ClassType(name) =>
        classScope.get(name.name) match {
          case Some(c) =>
            name.sym = c;
            IClassType(c)
          case None =>
            error(tree.pos, "Class " + name + " is unknown");
        }
    }

  // Analyze an expression
  private def analyzeExpr(varScope: VarScope, tree: Expr): Type =
    tree match {
      // ...  complter ...
      case NullLit() =>
        INullType

      case IntLit(_) =>
        IIntType
      // ...  complter ...
    }

  private def checkArgs(pos: Int, argpos: List[Int], argtypes: List[Type], paramtypes: List[Type]): Unit = {
    if (argtypes.length != paramtypes.length)
      Report.error(pos, "Wrong number of arguments: " + argtypes.length +
                   "; expected " + paramtypes.length);

    List.map3(argpos, argtypes, paramtypes)(Triple).foreach(
      trpl => {
        val Triple(pos, argtype, paramtype) = trpl;
        checkSubtype(pos, "Incompatible argument type", argtype, paramtype);
      }
    );
  }

  /**
   * Report an error
   */
  private def error(pos: Int, message: String): Type = {
    Report.error(pos, message);
    IBadType
  }

  /**
   * Generates an error if the expected type is not equal
   * to the found type.
   */
  private def checkSametype(pos: Int, found: Type, expected: Type): Unit =
    if (!found.isSametype(expected))
      typeError(pos, "Type mismatch", found, expected);

  /**
   * Generates an error if the found type is not equal to IIntType.
   */
  private def checkIntType(pos: Int, found: Type): Unit =
    checkSametype(pos, found, IIntType);

  /**
   * Generates an error if the expected type is not a super type
   * of the found type.
   */
  private def checkSubtype(pos: Int, header: String, found: Type, expected: Type) =
    if (!found.isSubtype(expected))
      typeError(pos, header, found, expected);

  /**
   * Generates an error message for unexpected types.
   */
  private def typeError(pos: Int, header: String, found: Type, expected: Type) =
    Report.error(pos, header + "\nexpected type: " + expected + "\n" +
                      "actual type  : " + found);

}
